using System;
using System.Collections.Generic;
using System.Windows.Input;

namespace InterpolationTest
{
    public class ClientSim
    {
        public bool isExtrapolating = false;
        public double renderTime = 0.0;
        public double clientTime = 0.0;
        public bool firstSnapshot = true;
        public bool smoothTimerState = true;

        public Vector2D localPos1 = new Vector2D(0, 0);
        public Vector2D localVel1 = Settings.defaultVelocity;
        public Vector2D localPos2 = new Vector2D(0, 0);
        public Vector2D localVel2 = Settings.defaultVelocity;

        public List<Snapshot> snapshots = new List<Snapshot>();
        public double[] measuredLatency = new double[Settings.latencySampleAmmount];
        public int measuredLatencyIndex;
        public double targetTime;
        public double renderTimeLast;

        public DateTime targetSetTime;
        private System.Diagnostics.Stopwatch swex = new System.Diagnostics.Stopwatch();
        private System.Diagnostics.Stopwatch swup = new System.Diagnostics.Stopwatch();

        private int findSnapshot(double t)
        {
            for (int i = 0; i < snapshots.Count - 1; ++i)
            {
                Snapshot snapshot0 = snapshots[i + 0];
                Snapshot snapshot1 = snapshots[i + 1];
                if (t >= snapshot0.m_time && t <= snapshot1.m_time)
                {
                    return i;
                }
            }

            return -1;
        }

        public void addSnapshot(double snapTime, Vector2D pos, Vector2D vel)
        {
            if (firstSnapshot)
            {
                firstSnapshot = false;

                renderTime = snapTime - Settings.timeOffset;
                targetTime = renderTime;
                localPos2 = pos;
                localVel2 = vel;
            }

            Snapshot snapshot = new Snapshot();
            snapshot.m_time = snapTime;
            snapshot.m_pos = pos;
            snapshot.m_vel = vel;
            snapshots.Add(snapshot);
            
            setTargetTime(snapTime);
        }

        private void setTargetTime(double snapTime)
        {
            //Measure Network Health
            measuredLatency[measuredLatencyIndex] = clientTime - snapTime;

            if (measuredLatencyIndex == Settings.latencySampleAmmount - 1)
            {
                measuredLatencyIndex = 0;
            }
            else
            {
                ++measuredLatencyIndex;
            }

            double maxLatency = 0;
            double minLatency = 999;
            double sumLatency = 0;
            for (int i = 0; i < Settings.latencySampleAmmount; ++i)
            {
                maxLatency = Math.Max(maxLatency, measuredLatency[i]);
                minLatency = Math.Min(minLatency, measuredLatency[i]);
                sumLatency += measuredLatency[i];
            }

            double avrLatency = sumLatency / Settings.latencySampleAmmount;

            double jitterRange = maxLatency - minLatency;

            // Calculate current targetTime

            double instantTargetTime = clientTime - avrLatency - jitterRange / 2 - Settings.timeOffset;
            double instantdifferenceTime = Math.Abs(renderTime - instantTargetTime);
            
            // Decide if targetTime update is needed
            if (swex.Elapsed.TotalMilliseconds / 1000 >= Settings.reactExtrapoTime || instantdifferenceTime > Settings.reactDifference)
            {
                targetTime = instantTargetTime;
                System.Diagnostics.Debug.WriteLine("setTargetTime: SET at clientTIme: " + clientTime);
            }
        }

        

        // Lerp that is framerate-independent.See https://www.scirra.com/blog/ashley/17/using-lerp-with-delta-time for the explanation and maths !
        public double lerpDT(double a, double b, double f, double dt)
        {
            double t = 1.0 - Math.Pow(f, dt);
            return ((a) * (1 - (t)) + (b) * (t));
        }

        public void update(double elapsed)
        {
            clientTime += elapsed;
            renderTime += elapsed;
            targetTime += elapsed;

            // process inputs
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.W))
            {
                localVel1.Y += elapsed;
            }
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.S))
            {
                localVel1.Y -= elapsed;
            }
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.A))
            {
                localVel1.X -= elapsed;
            }
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.D))
            {
                localVel1.X += elapsed;
            }

            if (smoothTimerState)
            {
                // speed up or slow down the display time to match the target time
                // that's pretty much the cause of the stuttering, however without that code, changes in latency mean the client and server timers are no longer synchronized properly
                // ie. try to PageUp to increase the latency to 1 second -> the client extrapolates all the time (green dot filled)

                // This section should be re-thought. It should limit the speed of the interpolation in such a way that time is not moving faster than physics and render can update.
                // Currently it is set up so the interpolation moves slower when it needs to interpolate more time distance.
                // Keep in mind that "setTargetTime" does not set the final "targetTime" the first time it runs. Due to it averaging several calclulated latencies the "targetTime" changes the more data it collects.
                double delta = Math.Abs(renderTime - targetTime);
                // Limit maximum interpolation speed to one second per second. By doing this local position doesn't need to be saved in case time moves backwards.
                // f*delta=renderToTargetDistance
                //delta = Math.Max(delta, (1.0 / delta));
                // Limit aboslut Maximum. No gegatives.
                delta = Math.Max(delta, 0.0);
                // Limit minimum interpolation speed. 1.0 = no change.
                delta = Math.Min(delta, 0.8);

                double proposedTime = lerpDT(renderTime, targetTime, delta, elapsed);
                // Limit maximum interpolation speed to one second per second when moving to a target in the past. By doing this local position doesn't need to be saved in case time moves backwards.
                if (proposedTime<renderTime-elapsed)
                {
                    // proposedTime tries to set back renderTime by more than one second per second.
                    // This stops render time for this cycle.
                    renderTime = renderTime - elapsed;
                }
                else
                {
                    renderTime = proposedTime;
                }
            }

            // simulate physics based on renderTime
            double elapsedRenderTime = renderTime - renderTimeLast;
            localPos1 = localPos1 + localVel1 * elapsedRenderTime; // move player 1 (that one is local only, no interpolation)
            localPos2 = localPos2 + localVel2 * elapsedRenderTime; // move player 2 (that's the one server-interpolated)

            // process local snapshots
            while (snapshots.Count > 0)
            {
                Snapshot first = snapshots[0];
                if (first.m_time < renderTime - Settings.snapStoreTime)
                {
                    // discard too old snapshots
                    snapshots.RemoveAt(0);
                }
                else
                {
                    break;
                }
            }

            //Render Network Targets
            int currentSnapshotIDrender = findSnapshot(renderTime);
            if (currentSnapshotIDrender == -1)
            {
                // extrapolate latest snapshot
                isExtrapolating = true;

                swex.Start();

                if (snapshots.Count > 0)
                {
                    Snapshot lastSnapshot = snapshots[snapshots.Count - 1];

                    double t = renderTime - lastSnapshot.m_time;
                    localPos2 = lastSnapshot.m_pos + lastSnapshot.m_vel * t;
                }
            }
            else
            {
                isExtrapolating = false;

                swex.Reset();

                Snapshot snapshot0 = snapshots[currentSnapshotIDrender + 0];
                Snapshot snapshot1 = snapshots[currentSnapshotIDrender + 1];
                
                // interpolate between current and next snapshot See this: https://gamedev.stackexchange.com/questions/12754/how-to-interpolate-between-two-game-states

                double elapsedSinceUpdate = renderTime - snapshot0.m_time;

                double tr = elapsedSinceUpdate / Settings.snapStep;
                tr = Math.Max(tr, 0.0);
                tr = Math.Min(tr, 1.0);

                localPos2 = snapshot0.m_pos + (snapshot1.m_pos - snapshot0.m_pos) * tr;
                localVel2 = snapshot0.m_vel + (snapshot1.m_vel - snapshot0.m_vel) * tr;
            }
            // Save renderTime of last render. Used to calculate elapsedRenderTime
            renderTimeLast = renderTime;
        }
    }
}
